#!/bin/sh

# Copyright (C) 2014 Assaf Gordon (assafgordon@gmail.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

##
## Runs the gnu-savannah-eval perl script
## on a given directory, then generates an HTML report for it.
##

die()
{
    BASE=$(basename "$0")
    echo "$BASE: error: $@" >&2
    exit 1
}

usage()
{
    BASE=$(basename "$0")
    echo "GNU-Savannah Evaluation - helper script
Copyright (C) 2014 A. Gordon (assafgordon@gmail.com)
License: GPLv3-or-later

Usage: $BASE [OPTIONS] PROJECT-NAME DIRECTORY

Will scan DIRECTORY with the gnu-sv-evaluation script,
and produce a Markdown and an HTML file, named based on PROJECT.

Options:

   -h              = show this help screen.
   --no-file-limit = Pass --no-file-limit to the perl evaluation script.
                     By default, evaluation stops after 4,000 files
                     (with a warning added to the output report).
Example:

    \$ $BASE my-project-0.1 \$HOME/projects/mycode

Will generate 'my-project-0.1.md' and 'my-project-0.1.html' .
"
    exit 0
}

NOLIMIT=
if [ "x$1" = "x--no-file-limit" ] ; then
  NOLIMIT=$1
  shift 1
fi
test "x$1" = "x-h" && usage

PROJECT=$1
DIRECTORY=$2

# Verify input parameters
test -z "$PROJECT" \
    && die "missing PROJECT parameter. Use -h for help."
test -z "$DIRECTORY" \
    && die "missing DIRECTORY parameter. Use -h for help."
echo "$PROJECT" | grep -qE '^[A-Za-z0-9\_ \-]+$' \
    || die "invalid PROJECT name '$PROJECT' (only alphanumeric chars allowed)"
test -d "$DIRECTORY" \
    || die "invalid directory '$DIRECTORY'"

# Verify required files
SCRIPTPATH=$(dirname $(readlink -f "$0")) \
    || die "failed to get script's directory"
EVAL_SCRIPT="$SCRIPTPATH/gsv-eval.pl"
test -x "$EVAL_SCRIPT" \
    || die "Perl script ($EVAL_SCRIPT) not found/not executable"
CSS_FILE="$SCRIPTPATH/gsv-eval.css"
test -e "$CSS_FILE" \
     || die "CSS file ($CSS_FILE) not found"

TIMESTAMP=$(date +%F-%H%M%S) || die "failed to get timestamp"

OUTPUT="$PROJECT.$TIMESTAMP"

"$EVAL_SCRIPT" $NOLIMIT --project "$PROJECT" "$DIRECTORY" > "$OUTPUT.md" \
    || die "evaluation script failed"

pandoc --from markdown \
       --to html \
       --table-of-contents \
       --include-in-header "$CSS_FILE" < "$OUTPUT.md" > "$OUTPUT.html" \
    || die "pandoc failed"

echo "$OUTPUT.html"
